@@ -1435,14 +1435,14 @@ ccio_init_resource(struct resource *res, char *name, unsigned long ioaddr)
 	res->flags = IORESOURCE_MEM;
 	res->start = (unsigned long)(signed) __raw_readl(ioaddr) << 16;
 	res->end = (unsigned long)(signed) (__raw_readl(ioaddr + 4) << 16) - 1;
-	if (res->end < res->start)
+	if (res->end + 1 == res->start)
 		return;
 	res->name = name;
 	result = request_resource(&iomem_resource, res);
 	if (result < 0) {
 		printk(KERN_ERR 
-		       "%s: failed to claim CCIO bus address space!\n", 
-		       __FILE__);
+		       "%s: failed to claim CCIO bus address space (%lx,%lx)\n", 
+		       __FILE__, res->start, res->end);
 	}
 }
 
@@ -1458,29 +1458,59 @@ static void __init ccio_init_resources(struct ioc *ioc)
 			(unsigned long)&ioc->ioc_hpa->io_io_low_hv);
 }
 
-#ifdef NASTY_HACK_FOR_K_CLASS
-void ccio_extend_mmio_range(struct ioc *ioc)
+static void expand_ioc_area(struct ioc *ioc, unsigned long size,
+		unsigned long min, unsigned long max, unsigned long align)
 {
+#ifdef NASTY_HACK_FOR_K_CLASS
 	__raw_writel(0xfffff600, (unsigned long)&(ioc->ioc_hpa->io_io_high));
 	ioc->mmio_region[0].end = 0xf5ffffff;
-}
 #endif
+}
 
-struct resource * ccio_find_parisc_resource(const struct parisc_device *dev)
+static struct resource *ccio_get_resource(struct ioc* ioc,
+		const struct parisc_device *dev)
 {
-	struct ioc *ioc = ccio_get_iommu(dev);
-	if (!ioc)
+	if (!ioc) {
 		return &iomem_resource;
-
-	if ((ioc->mmio_region->start <= dev->hpa) &&
+	} else if ((ioc->mmio_region->start <= dev->hpa) &&
 			(dev->hpa < ioc->mmio_region->end)) {
 		return ioc->mmio_region;
-	} else {
+	} else if (((ioc->mmio_region + 1)->start <= dev->hpa) &&
+			(dev->hpa < (ioc->mmio_region + 1)->end)) {
 		return ioc->mmio_region + 1;
+	} else {
+		return NULL;
 	}
 }
 
+int ccio_allocate_resource(const struct parisc_device *dev,
+		struct resource *res, unsigned long size,
+		unsigned long min, unsigned long max, unsigned long align,
+		void (*alignf)(void *, struct resource *, unsigned long, unsigned long),
+		void *alignf_data)
+{
+	struct ioc *ioc = ccio_get_iommu(dev);
+	struct resource *parent = ccio_get_resource(ioc, dev);
+	if (!parent)
+		return -EBUSY;
 
+	if (!allocate_resource(parent, res, size, min, max, align, alignf,
+			alignf_data))
+		return 0;
+
+	expand_ioc_area(ioc, size, min, max, align);
+	return allocate_resource(parent, res, size, min, max, align, alignf,
+			alignf_data);
+}
+
+int ccio_request_resource(const struct parisc_device *dev,
+		struct resource *res)
+{
+	struct ioc *ioc = ccio_get_iommu(dev);
+	struct resource *parent = ccio_get_resource(ioc, dev);
+
+	return request_resource(parent, res);
+}
 /**
  * ccio_probe - Determine if ccio should claim this device.
  * @dev: The device which has been found
