@@ -507,12 +507,11 @@ int pdc_get_initiator( struct hardware_path *hwpath, unsigned char *scsi_id,
 
 /* BCJ-XXXX series boxes. E.G. "9000/785/C3000" */
 #define IS_SPROCKETS() (strlen(boot_cpu_data.pdc.sys_model_name) == 14 && \
-	strncmp(boot_cpu_data.pdc.sys_model_name, "9000/785", 9) == 0)
+	strncmp(boot_cpu_data.pdc.sys_model_name, "9000/785", 8) == 0)
 
 	retval = mem_pdc_call(PDC_INITIATOR, PDC_GET_INITIATOR, 
 			      __pa(pdc_result), __pa(hwpath));
 
-
 	if (retval >= PDC_OK) {
 		*scsi_id = (unsigned char) pdc_result[0];
 
@@ -534,10 +533,12 @@ int pdc_get_initiator( struct hardware_path *hwpath, unsigned char *scsi_id,
 		** pdc_result[3]	PDC suggested SCSI rate
 		*/
 
+		/*
+		** XXX REVISIT: Doesn't look like PAT PDC does the same.
+		** Problem is A500 also exports 50-pin SE SCSI port.
+		*/
 		if (IS_SPROCKETS()) {
 			/*
-			** Revisit: PAT PDC do the same thing?
-			** A500 also exports 50-pin SE SCSI.
 			**	0 == 8-bit
 			**	1 == 16-bit
 			*/
@@ -691,43 +692,45 @@ int pdc_do_reset()
 
 /*
  * pdc_soft_power_info - Enable soft power switch.
- * @power_reg: The return buffer
+ * @power_reg: address of soft power register
  *
- * Enable the soft power switch and return the absolute address 
- * of the soft power switch register
+ * Return the absolute address of the soft power switch register
  */
 int __init pdc_soft_power_info(unsigned long *power_reg)
 {
 	int retval;
 
-	spin_lock_irq(&pdc_lock);
+	*power_reg = (unsigned long) (-1);
 	
+	spin_lock_irq(&pdc_lock);
 	retval = mem_pdc_call(PDC_SOFT_POWER, PDC_SOFT_POWER_INFO, __pa(pdc_result), 0);
 	if (retval == PDC_OK) {
-		*power_reg = pdc_result[0];
-		retval = mem_pdc_call(PDC_SOFT_POWER, PDC_SOFT_POWER_ENABLE, __pa(pdc_result), 1);
+                convert_to_wide(pdc_result);
+                *power_reg = f_extend(pdc_result[0]);
 	}
-	
 	spin_unlock_irq(&pdc_lock);
-	
-	if (retval != PDC_OK)
-		*power_reg = (unsigned long) (-1);
 
 	return retval;
 }
 
 /*
- * pdc_soft_power_shutdown - Shutdown system if possible.
+ * pdc_soft_power_button - Control the soft power button behaviour
+ * @sw_control: 0 for hardware control, 1 for software control 
  *
- * If supported by PDC this functions shuts down the system and
- * does not return to the caller. If not supported by the system
- * this function won't have any effect.
+ *
+ * This PDC function places the soft power button under software or
+ * hardware control.
+ * Under software control the OS may control to when to allow to shut 
+ * down the system. Under hardware control pressing the power button 
+ * powers off the system immediately.
  */
-void pdc_soft_power_shutdown(void)
+int pdc_soft_power_button(int sw_control)
 {
+	int retval;
 	spin_lock_irq(&pdc_lock);
-	mem_pdc_call(PDC_SOFT_POWER, PDC_SOFT_POWER_ENABLE, __pa(pdc_result), 0);
+	retval = mem_pdc_call(PDC_SOFT_POWER, PDC_SOFT_POWER_ENABLE, __pa(pdc_result), sw_control);
 	spin_unlock_irq(&pdc_lock);
+	return retval;
 }
 
 /*
