@@ -53,22 +53,38 @@ extern void ipi_interrupt(int, void *, struct pt_regs *);
 #endif /* DEBUG_IRQ */
 
 #define EIEM_MASK(irq)       (1UL<<(MAX_CPU_IRQ-IRQ_OFFSET(irq)))
-#define CLEAR_EIEM_BIT(irq)  set_eiem(get_eiem() & ~EIEM_MASK(irq))
-#define SET_EIEM_BIT(irq)    set_eiem(get_eiem() | EIEM_MASK(irq))
+
+/* Bits in EIEM correlate with cpu_irq_action[].
+** Numbered *Big Endian*! (ie bit 0 is MSB)
+*/
+static unsigned long cpu_eiem = 0;
 
 static spinlock_t irq_lock = SPIN_LOCK_UNLOCKED;  /* protect IRQ regions */
 
+#ifdef CONFIG_SMP
+static void cpu_set_eiem(void *info)
+{
+	set_eiem((unsigned long) info);
+}
+#endif
+
 static inline void disable_cpu_irq(void *unused, int irq)
 {
-	CLEAR_EIEM_BIT(irq);
+	unsigned long eirr_bit = EIEM_MASK(irq);
+
+	cpu_eiem &= ~eirr_bit;
+	set_eiem(cpu_eiem);
+        smp_call_function(cpu_set_eiem, (void *) cpu_eiem, 1, 1);
 }
 
 static void enable_cpu_irq(void *unused, int irq)
 {
-	unsigned long mask = EIEM_MASK(irq);
+	unsigned long eirr_bit = EIEM_MASK(irq);
 
-	mtctl(mask, 23);
-	SET_EIEM_BIT(irq);
+	mtctl(eirr_bit, 23);	/* clear EIRR bit before unmasking */
+	cpu_eiem |= eirr_bit;
+        smp_call_function(cpu_set_eiem, (void *) cpu_eiem, 1, 1);
+	set_eiem(cpu_eiem);
 }
 
 /* mask and disable are the same at the CPU level
@@ -78,10 +94,16 @@ static void enable_cpu_irq(void *unused, int irq)
 
 static inline void unmask_cpu_irq(void *unused, int irq)
 {
-	SET_EIEM_BIT(irq);
+	unsigned long eirr_bit = EIEM_MASK(irq);
+	cpu_eiem |= eirr_bit;
+	/* NOTE: sending an IPI will cause do_cpu_irq_mask() to
+	** handle *any* unmasked pending interrupts.
+	** ie We don't need to check for pending interrupts here.
+	*/
+        smp_call_function(cpu_set_eiem, (void *) cpu_eiem, 1, 1);
+	set_eiem(cpu_eiem);
 }
 
-
 static struct irqaction cpu_irq_actions[IRQ_PER_REGION] = {
 	[IRQ_OFFSET(TIMER_IRQ)] { handler: timer_interrupt, name: "timer", },
 #ifdef CONFIG_SMP
@@ -105,17 +127,18 @@ struct irq_region *irq_region[NR_IRQ_REGS] = {
 
 /*
 ** Generic interfaces that device drivers can use:
-**    mask_irq()	temporarily block IRQ
+**    mask_irq()	block IRQ
 **    unmask_irq()	re-enable IRQ and trigger if IRQ is pending
-**    disable_irq()	turn IRQ off - ie ignore it
-**    enable_irq()	turn IRQ on  - ie start using it
+**    disable_irq()	block IRQ
+**    enable_irq()	clear pending and re-enable IRQ
 */
 
 void mask_irq(int irq)
 {
 	struct irq_region *region;
 
-	DBG_IRQ(irq, ("mask_irq(%d) %d+%d\n", irq, IRQ_REGION(irq), IRQ_OFFSET(irq)));
+	DBG_IRQ(irq, ("mask_irq(%d) %d+%d eiem 0x%lx\n", irq,
+				IRQ_REGION(irq), IRQ_OFFSET(irq), cpu_eiem));
 	irq = irq_cannonicalize(irq);
 	region = irq_region[IRQ_REGION(irq)];
 	if (region->ops.mask_irq)
@@ -126,7 +149,8 @@ void unmask_irq(int irq)
 {
 	struct irq_region *region;
 
-	DBG_IRQ(irq, ("unmask_irq(%d) %d+%d\n", irq, IRQ_REGION(irq), IRQ_OFFSET(irq)));
+	DBG_IRQ(irq, ("unmask_irq(%d) %d+%d eiem 0x%lx\n", irq,
+				IRQ_REGION(irq), IRQ_OFFSET(irq), cpu_eiem));
 	irq = irq_cannonicalize(irq);
 	region = irq_region[IRQ_REGION(irq)];
 	if (region->ops.unmask_irq)
@@ -137,7 +161,8 @@ void disable_irq(int irq)
 {
 	struct irq_region *region;
 
-	DBG_IRQ(irq, ("disable_irq(%d) %d+%d\n", irq, IRQ_REGION(irq), IRQ_OFFSET(irq)));
+	DBG_IRQ(irq, ("disable_irq(%d) %d+%d eiem 0x%lx\n", irq,
+				IRQ_REGION(irq), IRQ_OFFSET(irq), cpu_eiem));
 	irq = irq_cannonicalize(irq);
 	region = irq_region[IRQ_REGION(irq)];
 	if (region->ops.disable_irq)
@@ -150,7 +175,8 @@ void enable_irq(int irq)
 {
 	struct irq_region *region;
 
-	DBG_IRQ(irq, ("enable_irq(%d) %d+%d\n", irq, IRQ_REGION(irq), IRQ_OFFSET(irq)));
+	DBG_IRQ(irq, ("enable_irq(%d) %d+%d eiem 0x%lx\n", irq,
+				IRQ_REGION(irq), IRQ_OFFSET(irq), cpu_eiem));
 	irq = irq_cannonicalize(irq);
 	region = irq_region[IRQ_REGION(irq)];
 
@@ -164,9 +190,14 @@ int get_irq_list(char *buf)
 {
 #ifdef CONFIG_PROC_FS
 	char *p = buf;
-	int regnr;
+	unsigned int regnr = 0;
+
+	p += sprintf(p, "     ");
+#ifdef CONFIG_SMP
+	for (; regnr < smp_num_cpus; regnr++)
+#endif
+		p += sprintf(p, "      CPU%d ", regnr);
 
-	p += sprintf(p, " IRQ      count         Region    ISR");
 #ifdef PARISC_IRQ_CR16_COUNTS
 	p += sprintf(p, "[min/avg/max] (CPU cycle counts)");
 #endif
@@ -177,7 +208,7 @@ int get_irq_list(char *buf)
 	*/
 	spin_lock(&irq_lock);
 	for (regnr = 0; regnr < NR_IRQ_REGS; regnr++) {
-	    int i;
+	    unsigned int i;
 	    struct irq_region *region = irq_region[regnr];
 
             if (!region || !region->action)
@@ -187,7 +218,7 @@ int get_irq_list(char *buf)
 		struct irqaction *action = &region->action[i];
 		unsigned int irq_no = IRQ_FROM_REGION(regnr) + i;
 #ifdef CONFIG_SMP
-		int j;
+		unsigned int j;
 #endif
 
 		if (!action->handler)
@@ -379,38 +410,53 @@ void do_irq(struct irqaction *action, int irq, struct pt_regs * regs)
 
 
 /* ONLY called from entry.S:intr_extint() */
-void do_cpu_irq_mask(unsigned long mask, struct irq_region *region, struct pt_regs *regs)
+void do_cpu_irq_mask(struct irq_region *region, struct pt_regs *regs)
 {
-	unsigned long bit;
-	unsigned long orig_eiem;
-	int irq;
-
-#ifdef DEBUG_IRQ
-	if (mask != (1L << MAX_CPU_IRQ))
-	    printk(KERN_DEBUG "do_irq_mask %08lx %p %p\n", mask, region, regs);
-#endif
+	unsigned long eirr_val;
+	unsigned int i=3;	/* limit time in interrupt context */
 
 	/*
-	 * do_cpu_irq_mask is called with the PSW_I bit off. we don't
-	 * enable it until we've masked (cleared the bits in the eiem)
-	 * the current set of interrupts we are processing.
+	 * PSW_I or EIEM bits cannot be enabled until after the
+	 * interrupts are processed.
+	 * timer_interrupt() assumes it won't get interrupted when it
+	 * holds the xtime_lock...an unmasked interrupt source could
+	 * interrupt and deadlock by trying to grab xtime_lock too.
+	 * Keeping PSW_I and EIEM disabled avoids this.
 	 */
+	set_eiem(0UL);	/* disable all extr interrupt for now */
 
-	orig_eiem = get_eiem();
-	set_eiem(orig_eiem & ~mask);
-	local_irq_enable();
-	for (bit = (1L<<MAX_CPU_IRQ), irq = 0; mask && bit; bit>>=1, irq++) {
-		int irq_num;
-		if (!(bit&mask))
-			continue;
+	/* 1) only process IRQs that are enabled/unmasked (cpu_eiem)
+	 * 2) We loop here on EIRR contents in order to avoid
+	 *    nested interrupts or having to take another interupt
+	 *    when we could have just handled it right away.
+	 * 3) Limit the number of times we loop to make sure other
+	 *    processing can occur.
+	 */
+	while ((eirr_val = (mfctl(23) & cpu_eiem)) && --i) {
+		unsigned long bit = (1UL<<MAX_CPU_IRQ);
+		unsigned int irq = 0;
 
-		mask &= ~bit;	/* clear bit in mask - can exit loop sooner */
-		irq_num = region->data.irqbase + irq;
-		do_irq(&region->action[irq], irq_num, regs);
+		mtctl(eirr_val, 23); /* reset bits we are going to process */
+
+#ifdef DEBUG_IRQ
+		if (eirr_val != (1UL << MAX_CPU_IRQ))
+			printk(KERN_DEBUG "do_cpu_irq_mask  %x\n", eirr_val);
+#endif
+
+		for (; eirr_val && bit; bit>>=1, irq++)
+		{
+			unsigned int irq_num;
+			if (!(bit&eirr_val))
+				continue;
+
+			/* clear bit in mask - can exit loop sooner */
+			eirr_val &= ~bit;
+
+			irq_num = region->data.irqbase + irq;
+			do_irq(&region->action[irq], irq_num, regs);
+		}
 	}
-	set_eiem(orig_eiem);
-
-	/* Leaving with PSW_I bit set */
+	set_eiem(cpu_eiem);
 }
 
 
@@ -418,7 +464,7 @@ void do_cpu_irq_mask(unsigned long mask, struct irq_region *region, struct pt_re
 void do_irq_mask(unsigned long mask, struct irq_region *region, struct pt_regs *regs)
 {
 	unsigned long bit;
-	int irq;
+	unsigned int irq;
 
 #ifdef DEBUG_IRQ
 	if (mask != (1L<<MAX_CPU_IRQ))
@@ -426,7 +472,7 @@ void do_irq_mask(unsigned long mask, struct irq_region *region, struct pt_regs *
 #endif
 
 	for (bit = (1L<<MAX_CPU_IRQ), irq = 0; mask && bit; bit>>=1, irq++) {
-		int irq_num;
+		unsigned int irq_num;
 		if (!(bit&mask))
 			continue;
 
@@ -778,8 +824,15 @@ int probe_irq_off(unsigned long val)
 void __init init_IRQ(void)
 {
 	local_irq_disable();	/* PARANOID - should already be disabled */
-	mtctl(-1L, 23);	/* EIRR : clear all pending interrupts */
-        set_eiem(-1L);	/* enable all EIR bits */
+	mtctl(-1L, 23);		/* EIRR : clear all pending external intr */
+#ifdef CONFIG_SMP
+	if (!cpu_eiem)
+		cpu_eiem = EIEM_MASK(IPI_IRQ) | EIEM_MASK(TIMER_IRQ);
+#else
+	cpu_eiem = EIEM_MASK(TIMER_IRQ);
+#endif
+        set_eiem(cpu_eiem);	/* EIEM : enable all external intr */
+
 }
 
 #ifdef CONFIG_PROC_FS
