@@ -405,9 +405,9 @@
 #define	writew_b2l	__raw_writew
 #define	writel_b2l	__raw_writel
 #define	readw_raw	__raw_readw
-#define	readl_raw	__raw_readl
+#define	readl_raw(a)	__raw_readl((unsigned long)(a))
 #define	writew_raw	__raw_writew
-#define	writel_raw	__raw_writel
+#define	writel_raw(v,a)	__raw_writel(v,(unsigned long)(a))
 #else	/* Other bid-endian */
 #define	readw_l2b	readw
 #define	readl_l2b	readl
