@@ -305,7 +305,6 @@ int setup_arg_pages(struct linux_binprm *bprm)
 	int i;
 	struct vm_area_struct *mpnt;
 	struct mm_struct *mm = current->mm;
-	int ret;
 
 #ifdef ARCH_STACK_GROWSUP
 	/* Move the argument and environment strings to the bottom of the
@@ -359,7 +358,6 @@ int setup_arg_pages(struct linux_binprm *bprm)
 	
 	down_write(&current->mm->mmap_sem);
 	{
-		struct vm_area_struct *vma;
 		mpnt->vm_mm = mm;
 #ifdef ARCH_STACK_GROWSUP
 		mpnt->vm_start = stack_base;
@@ -375,17 +373,7 @@ int setup_arg_pages(struct linux_binprm *bprm)
 		mpnt->vm_pgoff = 0;
 		mpnt->vm_file = NULL;
 		mpnt->vm_private_data = (void *) 0;
-		vma = find_vma(current->mm, mpnt->vm_start);
-		if (vma) {
-			up_write(&current->mm->mmap_sem);
-			kmem_cache_free(vm_area_cachep, mpnt);
-			return -ENOMEM;
-		}
-		if ((ret = insert_vm_struct(current->mm, mpnt))) {
-			up_write(&current->mm->mmap_sem);
-			kmem_cache_free(vm_area_cachep, mpnt);
-			return ret;
-		}
+		insert_vm_struct(mm, mpnt);
 		mm->total_vm = (mpnt->vm_end - mpnt->vm_start) >> PAGE_SHIFT;
 	} 
 
@@ -619,10 +607,8 @@ int flush_old_exec(struct linux_binprm * bprm)
 
 	current->sas_ss_sp = current->sas_ss_size = 0;
 
-	if (current->euid == current->uid && current->egid == current->gid) {
+	if (current->euid == current->uid && current->egid == current->gid)
 		current->mm->dumpable = 1;
-		current->task_dumpable = 1;
-	}
 	name = bprm->filename;
 	for (i=0; (ch = *(name++)) != '\0';) {
 		if (ch == '/')
@@ -1011,7 +997,7 @@ int do_coredump(long signr, struct pt_regs * regs)
 	binfmt = current->binfmt;
 	if (!binfmt || !binfmt->core_dump)
 		goto fail;
-	if (!is_dumpable(current))
+	if (!current->mm->dumpable)
 		goto fail;
 	current->mm->dumpable = 0;
 	if (current->rlim[RLIMIT_CORE].rlim_cur < binfmt->min_coredump)
