@@ -333,7 +333,6 @@ struct task_struct {
 	/* ??? */
 	unsigned long personality;
 	int did_exec:1;
-	unsigned task_dumpable:1;
 	pid_t pid;
 	pid_t pgrp;
 	pid_t tty_old_pgrp;
@@ -441,8 +440,6 @@ struct task_struct {
 #define PT_TRACESYSGOOD	0x00000008
 #define PT_PTRACE_CAP	0x00000010	/* ptracer can follow suid-exec */
 
-#define is_dumpable(tsk)	((tsk)->task_dumpable && (tsk)->mm->dumpable)
-
 /*
  * Limit the stack by to some sane default: root can always
  * increase this limit if needed..  8MB seems reasonable.
@@ -796,8 +793,6 @@ extern void FASTCALL(add_wait_queue(wait_queue_head_t *q, wait_queue_t * wait));
 extern void FASTCALL(add_wait_queue_exclusive(wait_queue_head_t *q, wait_queue_t * wait));
 extern void FASTCALL(remove_wait_queue(wait_queue_head_t *q, wait_queue_t * wait));
 
-extern long kernel_thread(int (*fn)(void *), void * arg, unsigned long flags);
-
 #define __wait_event(wq, condition) 					\
 do {									\
 	wait_queue_t __wait;						\
