@@ -113,6 +113,10 @@ extern void ipc_init(void);
 extern void parisc_init(void);
 #endif
 
+#ifdef CONFIG_PERFMON
+extern void perfmon_init(void);
+#endif
+
 /*
  * Boot command-line arguments
  */
@@ -599,6 +603,9 @@ asmlinkage void __init start_kernel(void)
 	kmem_cache_sizes_init();
 	pgtable_cache_init();
 
+#ifdef CONFIG_PERFMON
+	perfmon_init();
+#endif
 	mempages = num_physpages;
 
 	fork_init(mempages);
