@@ -503,15 +503,6 @@ sys_init_module(const char *name_user, struct module *mod_user)
 
 	/* On some machines it is necessary to do something here
 	   to make the I and D caches consistent.  */
-#ifdef __hppa__
-	/* I guess we have a problem here because copy_from_user() has left
-	 * the new module code in the data cache, so we need to flush it
-	 * through.  Our copy_from_user() used to do cache flushing, but 
-	 * it has been removed.  Puzzled as to why other archs might want
-	 * to flush icache at this point.  rhirst@linuxcare.com
-	 */
-	flush_data_cache();
-#endif
 	flush_icache_range((unsigned long)mod, (unsigned long)mod + mod->size);
 
 	/* Update module references.  */
