@@ -753,6 +753,11 @@ shmem_file_write(struct file *file,const char *buf,size_t count,loff_t *ppos)
 	long		status;
 	int		err;
 
+	if ((ssize_t) count < 0)
+		return -EINVAL;
+
+	if (!access_ok(VERIFY_READ, buf, count))
+		return -EFAULT;
 
 	down(&inode->i_sem);
 
